@@ -8,7 +8,7 @@ static inline char* _easyxs_sv_to_str (pTHX_ SV* sv, U8 is_utf8) {
 
     char *str = is_utf8 ? SvPVutf8_nolen(sv) : SvPVbyte_nolen(sv);
 
-    size_t len = strnlen(str, 1 + SvCUR(sv));
+    size_t len = strlen(str);
     if (len != SvCUR(sv)) {
         croak("Cannot convert scalar to C string (NUL byte detected, offset %" UVf ")", (UV) len);
     }
